from mancala.agents.minimax import MiniMaxAgent
from mancala.agents.base import BaseAgent
from mancala.agents.exact import ExactAgent
from mancala.agents.human import HumanAgent
from mancala.agents.max import MaxAgent
from mancala.agents.random import RandomAgent

ALL_AI_AGENTS = ["random", "exact", "max", "minimax"]


def init_agent(agent_type: str, id: int) -> BaseAgent:
    assert agent_type in (ALL_AI_AGENTS + ["human"])
    if agent_type == "human":
        return HumanAgent(id)
    elif agent_type == "random":
        return RandomAgent(id)
    elif agent_type == "max":
        return MaxAgent(id)
    elif agent_type == "minimax":
        return MiniMaxAgent(id)
    elif agent_type == "exact":
        return ExactAgent(id)
    else:
        raise ValueError


import random

from mancala.agents.base import BaseAgent
from mancala.state.base import BaseState


class RandomAgent(BaseAgent):
    """Agent with random choice policy"""

    def __init__(self, id: int, deterministic: bool = False, seed=42):
        self.id = id
        self._seed = seed
        self.deterministic = deterministic

    def policy(self, state: BaseState) -> int:
        """
        Make a move.

        Params
        ---
        state: mancala state object

        Returns
        ---
        action: int
        """
        if self.deterministic:
            random.seed(self._seed)
        return random.choice(state.legal_actions(state.current_player))


from typing import List, Union
import random
import sys

import numpy as np

from mancala.agents.base import BaseAgent
from mancala.state.base import BaseState


def minimax(state: BaseState, depth: int, maximizing_player_id: int) -> float:
    """
    MiniMax function
    """
    # Ref: https://en.wikipedia.org/wiki/Minimax
    return alphabeta(state, depth, maximizing_player_id, -float("inf"), float("inf"))


def alphabeta(
    state: BaseState, depth: int, maximizing_player_id: int, alpha: float, beta: float
) -> float:
    """
    MiniMax with alpha-beta pruning
    """
    # Ref: https://en.wikipedia.org/wiki/Alpha–beta_pruning
    if depth == 0 or state.is_terminal():
        return state.rewards_float(maximizing_player_id)

    if state.turn == maximizing_player_id:
        for act in state.legal_actions(state.turn):
            child = state.clone()
            child.proceed_action(act)
            alpha = max(
                alpha, alphabeta(child, depth - 1, maximizing_player_id, alpha, beta)
            )
            if alpha >= beta:
                break
        return alpha
    else:
        for act in state.legal_actions(state.turn):
            child = state.clone()
            child.proceed_action(act)
            beta = min(
                beta, alphabeta(child, depth - 1, maximizing_player_id, alpha, beta)
            )
            if alpha >= beta:
                break
        return beta


class MiniMaxAgent(BaseAgent):
    """
    Agent based on mini-max algorithm
    """

    def __init__(self, id: int, depth: int = 2):
        self.deterministic = False
        self._seed = 42
        self._depth = depth
        self.id = id

    def policy(self, state: BaseState) -> int:
        legal_actions = state.legal_actions(state.current_player)
        action_rewards = [
            minimax(state.clone().proceed_action(a), self._depth, state.current_player)
            for a in legal_actions
        ]
        # print(legal_actions)
        # print(action_rewards)
        max_reward = max(action_rewards)
        max_actions = [
            a for a, r in zip(legal_actions, action_rewards) if r == max_reward
        ]
        if self.deterministic:
            random.seed(self._seed)
        return random.choice(max_actions)


from dataclasses import dataclass


@dataclass
class Rule:
    multi_lap: bool = True
    capture_opposite: bool = True
    continue_on_point: bool = True
    pockets: int = 6
    initial_stones: int = 4
    stones_half: int = 6 * 4


import argparse

from pandas import DataFrame

from mancala.agents import ALL_AI_AGENTS, HumanAgent, RandomAgent
from mancala.arena import play_arena
from mancala.game import CLIGame
from mancala.mancala import MancalaEnv

parser = argparse.ArgumentParser(description="Mancala playable on cli")
subparsers = parser.add_subparsers(dest="command")

play_parser = subparsers.add_parser("play")
play_parser.add_argument(
    "--player0",
    type=str,
    default="human",
    help="Player that makes the first move",
    choices=ALL_AI_AGENTS + ["human"],
)
play_parser.add_argument(
    "--player1",
    type=str,
    default="random",
    help="Player that makes move the next turn",
    choices=ALL_AI_AGENTS + ["human"],
)

arena_parser = subparsers.add_parser("arena")
arena_parser.add_argument(
    "--num_games",
    type=int,
    default=100,
    help="How many times each pairs of agents should play together",
)
arena_parser.add_argument(
    "--only",
    type=str,
    default="",
    help=f"Explicitly select agents to compare from {ALL_AI_AGENTS}",
)


def cli():
    args = parser.parse_args()
    if args.command == "play":
        env = MancalaEnv([args.player0, args.player1])
        game = CLIGame(env)
        game.play_cli()
    elif args.command == "arena":
        agents = ALL_AI_AGENTS
        if args.only:
            agents = args.only.split(",")
        wins = play_arena(agents, args.num_games)
        print(DataFrame(wins))
    elif args.command == "train":
        pass



from abc import ABC
from builtins import NotImplementedError
from mancala.rule import Rule
from typing import List


class BaseState(ABC):
    action_choices: List[str]
    turn: int
    rule: Rule

    def __repr__(self):
        raise NotImplementedError

    def clone(self):
        raise NotImplementedError

    def __hash__(self):
        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseState):
            return NotImplemented
        raise NotImplementedError

    @property
    def current_player(self) -> int:
        raise NotImplementedError

    def legal_actions(self, turn: int) -> List[int]:
        raise NotImplementedError

    def proceed_action(self, act: int) -> None:
        raise NotImplementedError()

    @property
    def rewards(self) -> List[float]:
        raise NotImplementedError

    def rewards_float(self, receiver_player_id: int) -> float:
        raise NotImplementedError

    def is_terminal(self) -> bool:
        raise NotImplementedError



from mancala.agents.random import RandomAgent
from mancala.mancala import MancalaEnv


def main():
    env = MancalaEnv(["random", "random"])

    # Try N games
    N = 10
    for i in range(N):
        state = env.reset()
        total_reward = 0
        done = False

        while not done:
            action = env.current_agent.policy(state)
            next_state, reward, done = env.step(action)
            total_reward += reward
            state = next_state
            env.state = state

        print(f"Episode {i}: Agent gets {total_reward} reward.")


if __name__ == "__main__":
    main()


import sys
import time
from typing import List, Union

from mancala.agents import BaseAgent, HumanAgent, RandomAgent
from mancala.mancala import MancalaEnv, turn_names


class CLIGame(object):
    def __init__(self, env: MancalaEnv, silent=False):
        self.env = env
        self.silent = silent

    def _step(self) -> None:
        if not self.silent:
            print("turn:", self.env.current_agent)
        act = self.env.current_agent.policy(self.env.state)
        (next_state, reward, done) = self.env.step(act)
        self.env.state = next_state

    def play_cli(self) -> None:
        while not self.env.state._done:
            print(self.env.state)
            self.render_cli_board()
            self._step()
        print("END GAME")
        print(f"Winner: {self.env.state._winner}")
        self.render_cli_board()

    def render_cli_board(self) -> None:
        self.env.render()

    def play_silent(self) -> int:
        while self.env.state._winner is None:
            self._step()
        return self.env.state._winner


from collections import defaultdict
from pprint import pprint
from typing import DefaultDict, List
from tqdm import tqdm

from pandas import DataFrame
from mancala.agents import ALL_AI_AGENTS
from mancala.game import CLIGame
from mancala.mancala import MancalaEnv


def play_arena(agents: List[str] = ALL_AI_AGENTS, fights: int = 100):
    wins: DefaultDict[str, DefaultDict[str, float]] = defaultdict(
        lambda: defaultdict(float)
    )
    for agent_name0 in agents:
        for agent_name1 in agents:
            p1wins = 0
            for _ in tqdm(
                range(fights), desc=f"p0 {agent_name0} vs p1 {agent_name1}", leave=False
            ):
                env = MancalaEnv([agent_name0, agent_name1])
                game = CLIGame(env, silent=True)
                winner = game.play_silent()
                p1wins += winner
            wins["p0_" + agent_name0]["p1_" + agent_name1] = p1wins / fights * 100

    return wins


if __name__ == "__main__":
    wins = play_arena()
    print(DataFrame(wins))



import random

from mancala.agents.base import BaseAgent
from mancala.state.base import BaseState


class MaxAgent(BaseAgent):
    """Agent with random choice policy"""

    def __init__(self, id: int, deterministic: bool = False, seed=42):
        self.id = id
        self._seed = seed
        self.deterministic = deterministic

    @staticmethod
    def _score_of_action(act, state: BaseState) -> float:
        turn = state.current_player
        state.proceed_action(act)
        reward = state.rewards[turn]
        return reward

    def policy(self, state: BaseState) -> int:
        """
        Make a move.

        Params
        ---
        state: mancala state object

        Returns
        ---
        action: int
        """
        legal_actions = state.legal_actions(state.current_player)
        action_rewards = [
            MaxAgent._score_of_action(a, state.clone()) for a in legal_actions
        ]
        max_reward = max(action_rewards)
        max_actions = [
            a for a, r in zip(legal_actions, action_rewards) if r == max_reward
        ]
        if self.deterministic:
            random.seed(self._seed)
        return random.choice(max_actions)



from abc import ABC
from typing import List, Union

import numpy as np

from mancala.state.base import BaseState


class BaseAgent(ABC):
    """Abstract classs for Mancala agent"""

    id: int

    def __init__(self, id: int, actions: Union[List[int], np.ndarray]):
        self.id = id

    def __repr__(self):
        return f"<{self.__class__.__name__} id:{self.id}>"

    def policy(self, state: BaseState) -> int:
        raise NotImplementedError


import random

from mancala.agents.base import BaseAgent
from mancala.state.base import BaseState


class ValueIterationAgent(BaseAgent):
    """Agent with random choice policy"""

    def __init__(self, id: int, deterministic: bool = False, seed=42):
        # self.id = id
        # self._seed = seed
        # self.deterministic = deterministic
        pass

    def policy(self, state: BaseState) -> int:
        """
        Make a move.

        Params
        ---
        state: mancala state object

        Returns
        ---
        action: int
        """
        # if self.deterministic:
        #     random.seed(self._seed)
        # return random.choice(state.legal_actions(state.current_player))
        pass


# https://github.com/openai/gym/blob/master/docs/creating-environments.md
from __future__ import annotations

import random
import sys
import time
from dataclasses import dataclass
from typing import Dict, List, Tuple, Union

import gym
import numpy as np
from gym import Env, error, spaces, utils
from gym.utils import seeding

from mancala.agents import BaseAgent, init_agent
from mancala.state.base import BaseState
from mancala.rule import Rule


turn_names = ["player0", "player1"]


class MancalaState(BaseState):
    """
    Mancala State
    ---
    The board state and its utils
    """

    def __init__(
        self,
        board: Union[np.ndarray, None] = None,
        turn: int = 0,  # player: 0, ai: 1
    ):
        self.rule = Rule()
        if board is not None:
            assert board.shape == ((self.rule.pockets + 1) * 2,)
            self.board = board
        else:
            self.board = MancalaState.init_board(self.rule)
        self.turn = turn

        self.hand = 0

    @staticmethod
    def init_board(rule: Rule) -> np.ndarray:
        board = np.zeros(((rule.pockets + 1) * 2,), dtype=np.int32)
        # Player 1 side
        for i in range(0, rule.pockets):
            board[i] = rule.initial_stones
        # Player 2 side
        for i in range(rule.pockets + 1, rule.pockets * 2 + 1):
            board[i] = rule.initial_stones
        return board

    def legal_actions(self, turn: int) -> List[int]:
        if turn == 0:
            all_actions = list(self._player0_field_range)
        else:
            all_actions = list(self._player1_field_range)
        return self.filter_available_actions(all_actions)

    @property
    def current_player(self) -> int:
        return self.turn

    def __repr__(self):
        return f"<MancalaState: [{self.board}, {self.turn}]>"

    def clone(self) -> MancalaState:
        return MancalaState(board=self.board.copy(), turn=self.turn)

    def get_reward(self, turn: Union[int, None] = None) -> int:
        point = self.board[self._active_player_point_index]
        return point

    @property
    def rewards(self) -> List[float]:
        return [
            self.board[self._player0_point_index],
            self.board[self._player1_point_index],
        ]

    def rewards_float(self, receiver_player_id) -> float:
        if self._done and self._winner == receiver_player_id:
            return 1
        elif self._done:
            return -1
        else:
            if receiver_player_id == 0:
                return 0.01 * (self.rewards[0] - self.rewards[1])
            else:
                return 0.01 * (self.rewards[1] - self.rewards[0])

    def take_pocket(self, idx: int) -> None:
        """
        Params
        idx: index of the pocket to manipulate
        """
        self.hand += self.board[idx]
        self.board[idx] = 0

    def fill_pocket(self, idx: int, num: int = 1) -> None:
        """
        Params
        idx: index of the pocket to manipulate
        num: number of stones to fill in
        """
        assert self.hand > 0 and num <= self.hand
        # print(f"[DEBUG] Fill {num} into idx:{idx} pocket")
        self.board[idx] += num
        self.hand -= num

    def next_idx(self, idx: int) -> int:
        """
        Params
        idx :int: index to check

        Returns
        :int: the next index
        """
        next_idx = (idx + 1) % ((self.rule.pockets + 1) * 2)
        return next_idx

    def opposite_idx(self, idx: int) -> int:
        """
        Params
        idx :int: index to check

        Returns
        :int: the opposide field index
        """
        assert idx <= self.rule.pockets * 2
        return self.rule.pockets * 2 - idx

    @property
    def _player0_field_range(self):
        return range(0, self.rule.pockets)

    @property
    def _player1_field_range(self):
        return range(self.rule.pockets + 1, self.rule.pockets * 2 + 1)

    @property
    def _player0_point_index(self) -> int:
        return self.rule.pockets

    @property
    def _player1_point_index(self) -> int:
        return self.rule.pockets * 2 + 1

    @property
    def _active_player_point_index(self) -> int:
        return (
            self._player0_point_index if self.turn == 0 else self._player1_point_index
        )

    def is_current_sided_pointpocket(self, idx: int) -> bool:
        if self.turn == 0:
            return idx == self.rule.pockets
        else:
            return idx == self.rule.pockets * 2 + 1

    def is_current_sided_fieldpocket(self, idx: int) -> bool:
        if self.turn == 0:
            return 0 <= idx < self.rule.pockets
        else:
            return self.rule.pockets + 1 <= idx < self.rule.pockets * 2 + 1

    @property
    def sided_all_actions(self) -> List[int]:
        if self.turn == 0:
            return list(self._player0_field_range)
        else:
            return list(self._player1_field_range)

    def filter_available_actions(self, actions: List[int]) -> List[int]:
        return [i for i in actions if self.board[i] > 0]

    @property
    def sided_available_actions(self) -> List[int]:
        return self.filter_available_actions(self.sided_all_actions)

    @property
    def _winner(self) -> Union[int, None]:
        winner: Union[int, None] = None
        p0_actions = self.filter_available_actions(list(self._player0_field_range))
        p1_actions = self.filter_available_actions(list(self._player1_field_range))
        p0_points = self.board[self._player0_point_index]
        p1_points = self.board[self._player1_point_index]
        if len(p0_actions) == 0:
            p1_points += sum([self.board[i] for i in p1_actions])
        if len(p1_actions) == 0:
            p0_points += sum([self.board[i] for i in p0_actions])

        if p0_points > self.rule.stones_half:
            winner = 0
        elif p1_points > self.rule.stones_half:
            winner = 1
        elif len(p0_actions) == 0 or len(p1_actions) == 0:
            winner = 1 * (p1_points > p0_points)
        return winner

    @property
    def _done(self) -> bool:
        return self._winner is not None

    def is_terminal(self) -> bool:
        return self._done

    def proceed_action(self, idx: int):
        self.take_pocket(idx)
        continue_turn = False
        for _ in range(self.hand):
            idx = self.next_idx(idx)
            if (
                self.hand == 1
                and self.rule.continue_on_point
                and self.is_current_sided_pointpocket(idx)
            ):
                continue_turn = True
            if (
                self.hand == 1
                and self.rule.capture_opposite
                and self.is_current_sided_fieldpocket(idx)
                and self.board[idx] == 0
                and self.board[self.opposite_idx(idx)] > 0
            ):
                self.take_pocket(self.opposite_idx(idx))
                self.fill_pocket(self._active_player_point_index, self.hand)
                break
            self.fill_pocket(idx)
        if not (continue_turn and self.rule.multi_lap):
            self.flip_turn()
        return self

    def flip_turn(self):
        self.turn = 1 if self.turn == 0 else 0


class MancalaEnv(Env):
    metadata = {"render.modes": ["human"]}

    # Core Env functions
    # ------------------
    def __init__(self, agent_types: List[str]):
        super().__init__()
        self.rule = Rule()
        self.state = MancalaState()
        self.possible_agents = ["player0", "player1"]
        self.agents = self.init_agents(agent_types)

        # self.agents_dict = MancalaEnv.init_agents(agent_modes, agent_names=self.agents)
        # WIP
        # In respect to OpenSpiel API
        # self.agents = ["player0", "player1"]
        self.action_spaces = {
            i: spaces.Discrete(self.rule.pockets) for i in self.agents
        }
        self.observation_space = {
            i: spaces.Dict(
                {
                    "observation": spaces.Box(
                        low=0,
                        high=2 * self.rule.stones_half,
                        shape=(2, self.rule.pockets),
                        dtype=np.float16,
                    ),
                    "action_mask": spaces.Box(
                        low=0,
                        high=2 * self.rule.stones_half,
                        shape=(2 * self.rule.pockets,),
                        dtype=np.float16,
                    ),
                }
            )
            for i in self.agents
        }
        self.rewards = {i: 0 for i in self.agents}
        self.dones = {i: False for i in self.agents}
        self.infos = {
            i: {"legal_moves": list(range(0, self.rule.pockets))} for i in self.agents
        }
        # agent_selection

    def init_agents(self, agent_types):
        assert len(agent_types) == len(self.possible_agents)
        return [init_agent(atype, i) for i, atype in enumerate(agent_types)]

    @property
    def current_agent(self) -> BaseAgent:
        return self.agents[self.state.current_player]

    # @staticmethod
    # def init_agents(
    #     agent_modes: List[str], agent_names: List[str]
    # ) -> Dict[str, BaseAgent]:
    #     pass

    def reset(self) -> MancalaState:
        """
        Env core function
        """
        self.state = MancalaState()
        return self.state

    def step(self, action: int) -> Tuple[MancalaState, int, bool]:
        """
        Env core function
        """
        cloned_state = self.state.clone()
        cloned_state.proceed_action(action)
        reward = cloned_state.get_reward()
        done = cloned_state._done
        return (cloned_state, reward, done)

    def render(self, mode: str = "human") -> None:
        """
        Env core function
        ---
        Currently only supports CLI render mode.
        """
        print("\n" + "====" * (self.rule.pockets + 1))
        # AI side
        print(f"[{self.state.board[self.state._player1_point_index]:>2}]", end=" ")
        for i in self.state._player1_field_range[::-1]:
            print(f"{self.state.board[i]:>2}", end=" ")
        print("\n" + "----" * (self.rule.pockets + 1))
        # Player side
        print(" " * 4, end=" ")
        for i in self.state._player0_field_range:
            print(f"{self.state.board[i]:>2}", end=" ")
        print(f"[{self.state.board[self.state._player0_point_index]:>2}]", end=" ")
        print("\n" + "====" * (self.rule.pockets + 1))

    def close(self) -> None:
        """
        Env core function
        """
        pass

    # Common Env functions
    # --------------------
    @property
    def actions(self) -> List[int]:
        return list(range(self.rule.pockets))

    def transist(
        self, state: MancalaState, action: int
    ) -> Tuple[MancalaState, int, bool]:
        """
        Returns
        next_state:
        reward:
        done:
        """
        pass

    @staticmethod
    def transistion_func(state: BaseState, action: int) -> None:
        """
        Params
        state:
        action:

        Returns
        transition_probs: List[flaot]
        """
        pass



from mancala.agents.minimax import minimax
import random
import sys
from typing import List, Union

import numpy as np

from mancala.agents.base import BaseAgent
from mancala.state.base import BaseState


class HumanAgent(BaseAgent):
    """Agent with random choice policy"""

    def __init__(self, id: int, actions: Union[List[int], np.ndarray, None] = None):
        # super().__init__(actions)
        self.id = id
        self.hint = True

    def policy(self, state: BaseState) -> int:
        """
        Make a move.

        Params
        ---
        state: mancala state object

        Returns
        ---
        action: int
        """
        act = self.get_player_action(state)
        return act

    def get_player_action(self, state: BaseState) -> int:
        # Render action choices
        print(" " * 4, end=" ")
        action_choices = [str(i) for i in range(1, state.rule.pockets + 1)]
        for i in action_choices:
            print(f"{i:>2}", end=" ")
        print()
        if self.hint:
            for act in range(0, 6):
                v = minimax(state.clone().proceed_action(act), 4, 0)
                print(f"{v}", end=" ")
            print()

        # Receive human input
        while True:
            key_input = input("Take from a pocket (chose the index) \n> ")
            if key_input == "q":
                sys.exit()
            idx = action_choices.index(key_input)
            if self.id == 1:
                idx = state.rule.pockets * 2 - idx
            assert idx >= 0
            if idx in state.legal_actions(state.current_player):
                return idx
            else:
                print("Cannot pick from empty pocket:", idx)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mancala', 'mancala.agents', 'mancala.state']

package_data = \
{'': ['*']}

install_requires = \
['gym>=0.18.0,<0.19.0',
 'numpy>=1.20.1,<2.0.0',
 'pandas>=1.2.4,<2.0.0',
 'tqdm>=4.60.0,<5.0.0']

entry_points = \
{'console_scripts': ['mancala = mancala.cli:cli']}

setup_kwargs = {
    'name': 'mancala',
    'version': '0.2.1',
    'description': 'Mancala written in Python, playable in CLI (GUI coming soon)!',
    'long_description': "# Mancala\n\nMancala board game written in python.\n\n![img](https://github.com/qqhann/Mancala/blob/main/assets/preview_cli.png)\n\n## Features & Road maps\n\n- [x] Mancala playable on CLI\n- [x] Cmpatible with the gym API\n- [ ] Can train RL agents\n- [ ] Mancala playable on GUI\n\n## Installation\n\n```shell\n$ pip install mancala\n```\n\n## Usage\n\n### Play a game with agents\n\n```shell\n$ mancala play --player0 human --player1 random\n```\n\n### Compare each agents and plot their win rates\n\nThe values are player0's (first move) win rates in percentage\n\n```shell\n$ mancala arena\n            p0_random  p0_exact  p0_max  p0_minimax\np1_random        40.0      12.0     2.0         0.0\np1_exact         76.0      38.0    29.0         2.0\np1_max           91.0      55.0    27.0         1.0\np1_minimax       99.0      87.0    81.0        28.0\n```\n\n## Algorithms\n\nMancala is a game with perfect information.\nマンカラは完全情報ゲームです。\n\n### Value Iteration\n\nUsing Dynamic Programming (DP), calculate value for states and memorize them.\nUse the value to plan future actions.\n\nOther implementations\n\n- OpenSpiel value_iteration\n  algorithm <https://github.com/deepmind/open_spiel/blob/master/open_spiel/python/algorithms/value_iteration.py>\n  example <https://github.com/deepmind/open_spiel/blob/master/open_spiel/python/examples/value_iteration.py>\n\n### Policy Iteration\n\nUsing Dynamic Programming (DP), calculate value for states and memorize them.\nUse the value and policy for planning.\n\n## References\n\n- <https://github.com/mdavolio/mancala>\n\n### Multi agent RL\n\n- <https://github.com/deepmind/open_spiel>\n- <https://github.com/PettingZoo-Team/PettingZoo>\n",
    'author': 'Qiushi Pan',
    'author_email': 'qiu.gits@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/qqhann/Mancala',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import random

from mancala.agents.base import BaseAgent
from mancala.state.base import BaseState


class ExactAgent(BaseAgent):
    """Agent to use continue-on-point rule"""

    def __init__(self, id: int, deterministic: bool = False, seed=42):
        self.id = id
        self._seed = seed
        self.deterministic = deterministic

    @staticmethod
    def _score_of_action(act, state: BaseState) -> float:
        turn = state.current_player
        state.proceed_action(act)
        reward = state.rewards[turn]
        return reward

    @staticmethod
    def _turn_kept_by_action(state: BaseState, act: int) -> bool:
        turn = state.current_player
        state.proceed_action(act)
        kept = state.current_player == turn
        return kept

    def policy(self, state: BaseState) -> int:
        """
        Make a move.

        Params
        ---
        state: mancala state object

        Returns
        ---
        action: int
        """
        legal_actions = state.legal_actions(state.current_player)
        action_turn_kepts = [
            ExactAgent._turn_kept_by_action(state.clone(), a) for a in legal_actions
        ]
        exact_actions = [a for a, kept in zip(legal_actions, action_turn_kepts) if kept]
        if self.deterministic:
            random.seed(self._seed)
        if len(exact_actions) > 0:
            return random.choice(exact_actions)
        else:
            # action_rewards = [
            #     ExactAgent._score_of_action(a, state.clone()) for a in legal_actions
            # ]
            return random.choice(legal_actions)
